import tensorflow as tf
from tensorflow.keras import backend as K


def weighted_accuracy(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
    if len(tf.shape(y_true)) > 1:
        y_true = tf.argmax(y_true, axis=-1)
    if len(tf.shape(y_pred)) > 1:
        y_pred = tf.argmax(y_pred, axis=-1)
    y_true = tf.cast(y_true, tf.float32)
    y_pred = tf.cast(y_pred, tf.float32)

    correct_samples = K.cast(K.equal(y_true, y_pred), tf.float32)
    return K.mean(correct_samples)


def unweighted_accuracy(y_true: tf.Tensor, y_pred: tf.Tensor, n_classes: int = 4) -> tf.Tensor:
    cm = compute_confusion_matrix(y_true, y_pred)
    cm = normalize_confusion_matrix(cm)
    return tf.reduce_mean(tf.linalg.diag_part(cm))


def get_emotion_indices(y: tf.Tensor, emotion_index: int) -> tf.Tensor:
    return tf.reshape(tf.where(tf.equal(y, emotion_index)), (-1,))


def compute_confusion_matrix(y_true: tf.Tensor, y_pred: tf.Tensor) -> tf.Tensor:
    if len(tf.shape(y_true)) > 1:
        y_true = tf.argmax(y_true, axis=-1)
    if len(tf.shape(y_pred)) > 1:
        y_pred = tf.argmax(y_pred, axis=-1)
    return tf.math.confusion_matrix(y_true, y_pred, dtype=tf.float32)


def normalize_confusion_matrix(confusion_matrix: tf.Tensor, axis: int = 1) -> tf.Tensor:
    if axis == 0:
        return confusion_matrix / tf.reduce_sum(confusion_matrix, axis=0)
    elif axis == 1:
        cm_transpose = tf.transpose(confusion_matrix, (1, 0))
        norm_cm_transpose = cm_transpose / tf.reduce_sum(confusion_matrix, axis=1)
        return tf.transpose(norm_cm_transpose, (1, 0))
    else:
        raise ValueError("Invalid `axis` argument. Only 0 or 1 available")



from typing import Dict, Any

from tensorflow.keras.layers import Conv1D, BatchNormalization, LSTM, Bidirectional, Dense, Masking
import tensorflow as tf

from .base_model import BaseModel
from .layers.cnn_blocks import CNN1DBlock


def get_cnn(filters: int, kernel_size: int, stride: int = 1, activation: str = 'relu'):
    return tf.keras.Sequential([
        Conv1D(filters=filters, kernel_size=kernel_size, strides=stride, activation=activation),
        BatchNormalization()
    ])


class CNN1DLSTM(BaseModel):
    def __init__(
            self,
            config: Dict[str, Any] = None,
            **kwargs):
        super().__init__(config, **kwargs)

        if config is None:
            config = {}
        self.n_classes = config.get('n_classes', 4)
        self.num_channels = config.get('num_channels', [64, 64, 128, 128])
        self.kernel_sizes = config.get('kernel_sizes', [4, 3, 3, 3])
        self.pool_sizes = config.get('pool_sizes', [2, 2, 2, 2])
        self.bidirectional = config.get('bidirectional', True)
        self.rnn_units = config.get('rnn_units', [128])
        if isinstance(self.rnn_units, int):
            self.rnn_units = [self.rnn_units]
        self.norm_method = config.get('norm_method', 'layer')
        self.dropout = config.get('dropout', 0.)

        # CNN layers
        assert len(self.num_channels) == len(self.kernel_sizes) == len(
            self.pool_sizes), \
            f'Number of channel / kernel / pool sizes mismatch: {len(self.num_channels)}, ' \
            f'{len(self.kernel_sizes)} and {len(self.pool_sizes)}'

        self.cnn_layers = [CNN1DBlock(
            n_channel=c,
            kernel_size=k,
            pool_size=p,
            norm_method=self.norm_method
        ) for c, k, p in zip(self.num_channels, self.kernel_sizes, self.pool_sizes)]

        # RNN layer
        self.recurrent_layers = [
            LSTM(unit, return_sequences=True) if i != len(self.rnn_units)-1 else LSTM(unit)
            for i, unit in enumerate(self.rnn_units)
        ]

        if self.bidirectional:
            self.recurrent_layers = [Bidirectional(layer) for layer in self.recurrent_layers]
        # logits layer
        self.logits = Dense(self.n_classes)

    def get_config(self):
        return {'num_channels': self.num_channels,
                'kernel_sizes': self.kernel_sizes,
                'pool_sizes': self.pool_sizes,
                'bidirectional': self.bidirectional,
                'rnn_units': self.rnn_units,
                'n_classes': self.n_classes,
                'dropout': self.dropout}

    def call(self, x, training=False, **kwargs) -> tf.Tensor:
        # cnn layers
        for cnn_block in self.cnn_layers:
            x = cnn_block(x, training=training)

        # masking
        x = Masking()(x)

        # lstm
        for rnn in self.recurrent_layers:
            x = rnn(x)

        # dense
        if training:
            x = tf.nn.dropout(x, rate=self.dropout)
        x = self.logits(x)
        return tf.nn.softmax(x, axis=-1)


class TestModel(BaseModel):

    def __init__(self, config: Dict[str, Any], **kwargs):
        super().__init__(config, **kwargs)
        self.n_classes = config.get('n_classes', 4)
        self.logits = Dense(self.n_classes)

    def call(self, x, **kwargs):
        x = tf.reduce_mean(x, axis=1)
        x = self.logits(x)
        x = tf.nn.softmax(x, axis=-1)
        return x

    def get_config(self):
        super().get_config()



from typing import List, Union

import tensorflow as tf
import numpy as np

from .features.featureloader import FeatureLoader
from .features.preprocessing import load_waveform

AUTOTUNE = tf.data.experimental.AUTOTUNE
EMOTIONS = ['neutral', 'anger', 'happiness', 'sadness']


class DataLoader:
    def __init__(self,
                 csv_paths: Union[List[str], str],
                 feature_loader: FeatureLoader,
                 shuffle: bool = True):
        if isinstance(csv_paths, str):
            csv_paths = [csv_paths]
        self.csv_paths = csv_paths
        self.shuffle = shuffle
        self.feature_loader = feature_loader
        self.steps_per_epoch = None

    def read_csv(self) -> np.ndarray:
        lines = list()
        for csv in self.csv_paths:
            print(f"Reading {csv}...")
            with tf.io.gfile.GFile(csv, 'r') as f:
                lines += f.read().splitlines()[1:]  # skip header
        lines = np.array([line.split(',') for line in lines])  # line i: audio_path, emotion
        if self.shuffle:
            np.random.shuffle(lines)
        return lines

    def get_dataset(self, batch_size) -> Union[tf.data.Dataset, None]:
        data = self.read_csv()
        if len(data) == 0:
            return None
        dataset = tf.data.Dataset.from_tensor_slices(data)
        return self.process_dataset(dataset, batch_size=batch_size)

    @tf.function
    def preprocess(self, data):
        audio_path = data[0]
        emotion = data[1]
        with tf.device('/CPU:0'):
            waveform = load_waveform(audio_path)

            # apply augmentation and feature extraction
            features = self.feature_loader.extract(waveform)

            label = tf.argmax(tf.stack([tf.equal(emotion, e) for e in EMOTIONS], axis=-1))
            return features, label

    @tf.function
    def parse(self, data):
        return tf.numpy_function(
            self.preprocess,
            inp=[data[0], data[1]],
            Tout=[tf.float32, tf.int64]
        )

    def process_dataset(self,
                        dataset: Union[tf.data.Dataset, tf.data.TFRecordDataset],
                        batch_size: int) -> tf.data.Dataset:
        dataset = dataset.map(self.preprocess, num_parallel_calls=AUTOTUNE)

        if self.shuffle:
            dataset = dataset.shuffle(16, reshuffle_each_iteration=True)

        dataset = dataset.padded_batch(
            batch_size=batch_size,
            padded_shapes=(
                tf.TensorShape(self.feature_loader.shape),
                tf.TensorShape([])
            ),
            padding_values=(0., None),
            drop_remainder=False
        )

        dataset = dataset.prefetch(AUTOTUNE)
        self.steps_per_epoch = len(dataset)
        return dataset.repeat(None)
